r"""
Converts ECLab MPT files into an echemdb datapackage compatible CSV object.
"""
# ********************************************************************
#  This file is part of echemdb-converters.
#
#        Copyright (C) 2022 Albert Engstfeld
#        Copyright (C) 2022 Johannes Hermann
#        Copyright (C) 2022 Julian Rüth
#
#  echemdb-converters is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  echemdb-converters is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with echemdb-converters. If not, see <https://www.gnu.org/licenses/>.
# ********************************************************************

from echemdbconverters.ec_unit_package_adapter import ECUnitPackageAdapter


class ECLabAdapter(ECUnitPackageAdapter):
    r"""
    Some description.

    EXAMPLES::

        >>> from io import StringIO
        >>> file = StringIO('''EC-Lab ASCII FILE
        ... Nb header lines : 6
        ...
        ... Device metadata : some metadata
        ...
        ... mode\ttime/s\tEwe/V\t<I>/mA\tcontrol/V
        ... 1\t2\t3\t4\t5
        ... 1\t2.1\t3.1\t4.1\t5.1
        ... ''')
        >>> device = 'eclab'
        >>> from echemdbconverters.csvloader import CSVloader
        >>> from echemdbconverters.ec_unit_package_adapter import ECUnitPackageAdapter
        >>> ec = ECUnitPackageAdapter.create(device=device)(CSVloader.create(device)(file))
        >>> ec.loader.df
           mode  time/s  Ewe/V  <I>/mA  control/V
        0     1     2.0    3.0     4.0        5.0
        1     1     2.1    3.1     4.1        5.1

        >>> ec.df
           mode    t    E    I  control/V
        0     1  2.0  3.0  4.0        5.0
        1     1  2.1  3.1  4.1        5.1

        >>> ec.fields()  # doctest: +NORMALIZE_WHITESPACE
        [{'name': 'mode', 'type': 'integer'},
        {'name': 't', 'type': 'number', 'description': 'relative time', 'unit': 's', 'dimension': 't', 'original name': 'time/s'},
        {'name': 'E', 'type': 'number', 'description': 'working electrode potential', 'unit': 'V', 'dimension': 'E', 'original name': 'Ewe/V'},
        {'name': 'I', 'type': 'number', 'description': 'working electrode current', 'unit': 'mA', 'dimension': 'I', 'original name': '<I>/mA'},
        {'name': 'control/V', 'type': 'number', 'description': 'control voltage', 'unit': 'V', 'dimension': 'E'}]

    """

    @property
    def field_name_conversion(self):
        return {
            "time/s": "t",
            "Ewe/V": "E",
            "<I>/mA": "I",
        }



r"""
Creates standardized echemdb datapackage compatible CSV.

The file loaded must have the columns t, U/E, and I/j.
In addition, the fields to these columns must have a unit.

EXAMPLES::

    >>> from io import StringIO
    >>> file = StringIO(r'''t,E,j,x
    ... 0,0,0,0
    ... 1,1,1,1''')
    >>> from echemdbconverters.csvloader import CSVloader
    >>> metadata = {'figure description': {'schema': {'fields': [{'name':'t', 'unit':'s'},{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name':'x', 'unit':'m'}]}}}
    >>> ec = ECUnitPackageAdapter(CSVloader(file=file), fields=metadata['figure description']['schema']['fields'])
    >>> ec.df
       t  E  j  x
    0  0  0  0  0
    1  1  1  1  1

The original dataframe can be accessed from the loader::

    >>> ec.loader.df
       t  E  j  x
    0  0  0  0  0
    1  1  1  1  1

"""
# ********************************************************************
#  This file is part of echemdb-converters.
#
#        Copyright (C) 2022 Albert Engstfeld
#        Copyright (C) 2022 Johannes Hermann
#        Copyright (C) 2022 Julian Rüth
#
#  echemdb-converters is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  echemdb-converters is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with echemdb-converters. If not, see <https://www.gnu.org/licenses/>.
# ********************************************************************


import logging

logger = logging.getLogger("ECUnitPackageAdapter")


class ECUnitPackageAdapter:
    r"""
    Creates standardized echemdb datapackage compatible CSV.

    The file loaded must have the columns t, U or E, and I or j.
    In addition, the fields to these columns must have a unit.

    EXAMPLES::

        >>> from io import StringIO
        >>> file = StringIO(r'''t,E,j,x
        ... 0,0,0,0
        ... 1,1,1,1''')
        >>> from echemdbconverters.csvloader import CSVloader
        >>> metadata = {'figure description': {'schema': {'fields': [{'name':'t', 'unit':'s'},{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name':'x', 'unit':'m'}]}}}
        >>> ec = ECUnitPackageAdapter(CSVloader(file=file), fields=metadata['figure description']['schema']['fields'])
        >>> ec.df
           t  E  j  x
        0  0  0  0  0
        1  1  1  1  1

    A list of names describing the columns::

        >>> ec.column_names
        ['t', 'E', 'j', 'x']

        >>> ec.fields()
        [{'name': 't', 'type': 'integer', 'unit': 's'},
        {'name': 'E', 'type': 'integer', 'unit': 'V', 'reference': 'RHE'},
        {'name': 'j', 'type': 'integer', 'unit': 'uA / cm2'},
        {'name': 'x', 'type': 'integer', 'unit': 'm'}]

    """
    core_dimensions = {"time": ["t"], "voltage": ["E", "U"], "current": ["I", "j"]}

    def __init__(self, loader, fields=None, metadata=None):
        self.loader = loader
        self._fields = self.loader.derive_fields(fields=fields)
        self._metadata = metadata

    @staticmethod
    def create(device=None):
        r"""
        Calls a specific `EC adapter` based on a given device.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j,x
            ... 0,0,0,0
            ... 1,1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> metadata = {'figure description': {'schema': {'fields': [{'name':'t', 'unit':'s'},{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name':'x', 'unit':'m'}]}}}
            >>> ec = ECUnitPackageAdapter.create('eclab')(CSVloader(file=file), fields=metadata['figure description']['schema']['fields'])
            >>> ec.df
            t  E  j  x
            0  0  0  0  0
            1  1  1  1  1

        """
        if device == "eclab":
            from echemdbconverters.eclab_adapter import ECLabAdapter

            return ECLabAdapter

        raise KeyError(
            f"Device wth name '{device}' is unknown to the ECUnitPackageAdapter."
        )

    @property
    def field_name_conversion(self):
        """
        A dictionary which defines new names for column names of the loaded CSV.
        For example the loaded CSV could contain a column with name `time/s`.
        In the converted CSV that column should be named `t` instead.
        In that case {'time/s':'t'} should be returned.
        The property should be adapted in the respective device converters.


        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j,x
            ... 0,0,0,0
            ... 1,1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> ec = ECUnitPackageAdapter(CSVloader(file))
            >>> ec.field_name_conversion
            {}
        """
        return {}

    def fields(self):
        r"""
        A frictionless `Schema` object, including a `Fields` object
        describing the columns of the converted electrochemical data.

        In case the field names were not changed in property:name_conversion:
        the resulting schema is identical to that of the loader.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j,x
            ... 0,0,0,0
            ... 1,1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> metadata = {'figure description': {'schema': {'fields': [{'name':'t', 'unit':'s'},{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name':'x', 'unit':'m'}]}}}
            >>> ec = ECUnitPackageAdapter(CSVloader(file=file), fields=metadata['figure description']['schema']['fields'])
            >>> ec.fields() # doctest: +NORMALIZE_WHITESPACE
            [{'name': 't', 'type': 'integer', 'unit': 's'},
            {'name': 'E', 'type': 'integer', 'unit': 'V', 'reference': 'RHE'},
            {'name': 'j', 'type': 'integer', 'unit': 'uA / cm2'},
            {'name': 'x', 'type': 'integer', 'unit': 'm'}]

        A CSV with incomplete field information.::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j,x
            ... 0,0,0,0
            ... 1,1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> metadata = {'figure description': {'schema': {'fields': [{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name':'t', 'unit':'s'}]}}}
            >>> ec = ECUnitPackageAdapter(CSVloader(file=file), fields=metadata['figure description']['schema']['fields'])
            >>> ec.fields() # doctest: +NORMALIZE_WHITESPACE
            [{'name': 't', 'type': 'integer', 'unit': 's'},
            {'name': 'E', 'type': 'integer', 'unit': 'V', 'reference': 'RHE'},
            {'name': 'j', 'type': 'integer', 'unit': 'uA / cm2'},
            {'name': 'x', 'type': 'integer'}]

        A CSV with a missing potential axis which is, however, defined in the field description.::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,j,x
            ... 0,0,0
            ... 1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> metadata = {'figure description': {'schema': {'fields': [{'name':'t', 'unit':'s'},{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'}]}}}
            >>> ec = ECUnitPackageAdapter(CSVloader(file=file), fields=metadata['figure description']['schema']['fields'])
            >>> ec.fields()
            Traceback (most recent call last):
            ...
            KeyError: "No column with a 'voltage' axis."


        Fields with missing units in one o the core dimension fields.::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j,x
            ... 0,0,0,0
            ... 1,1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> metadata = {'figure description': {'schema': {'fields': [{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name':'t'}]}}}
            >>> ec = ECUnitPackageAdapter(CSVloader(file=file), fields=metadata['figure description']['schema']['fields'])
            >>> ec.fields()
            Traceback (most recent call last):
            ...
            KeyError: "No unit associated with the field named 't'"

        """
        import itertools

        from frictionless import Schema

        schema = Schema(fields=self._fields)

        for name in schema.field_names:
            # Change the name of specific fields.
            if name in self.field_name_conversion:
                schema.update_field(name, {"original name": name})
                schema.update_field(name, {"name": self.field_name_conversion[name]})
            # Validate that each field with a core dimension has a unit.
            if name in list(
                itertools.chain.from_iterable(list(self.core_dimensions.values()))
            ):
                try:
                    schema.get_field(name).custom["unit"]
                except KeyError as exc:
                    raise KeyError(
                        f"No unit associated with the field named '{name}'"
                    ) from exc

        # Validates that the column names contain a time, voltage and current axis.
        for key, item in self.core_dimensions.items():
            if not set(item).intersection(set(schema.field_names)):
                raise KeyError(f"No column with a '{key}' axis.")

        return schema.fields

    @property
    def column_names(self):
        """
        The EC file must have at least three dimensions, including time, voltage and current.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j,x
            ... 0,0,0,0
            ... 1,1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> metadata = {'figure description': {'schema': {'fields': [{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name':'x', 'unit':'m'},{'name':'t', 'unit':'s'}]}}}
            >>> ec = ECUnitPackageAdapter(CSVloader(file), fields=metadata['figure description']['schema']['fields'])
            >>> ec.column_names
            ['t', 'E', 'j', 'x']

        """
        from frictionless import Schema

        return Schema(fields=self.fields()).field_names

    @property
    def df(self):
        """
        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j,x
            ... 0,0,0,0
            ... 1,1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> metadata = {'figure description': {'schema': {'fields': [{'name':'t', 'unit':'s'},{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name':'x', 'unit':'m'}]}}}
            >>> ec = ECUnitPackageAdapter(CSVloader(file), fields=metadata['figure description']['schema']['fields'])
            >>> ec.df
               t  E  j  x
            0  0  0  0  0
            1  1  1  1  1

        """
        df = self.loader.df.copy()
        df.columns = self.column_names
        return df

    @property
    def metadata(self):
        r"""
        Returns metadata associated with the CSV.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j,x
            ... 0,0,0,0
            ... 1,1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> metadata = {'figure description': {'fields': [{'name':'t', 'unit':'s'},{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name':'x', 'unit':'m'}]}}
            >>> ec = ECUnitPackageAdapter(CSVloader(file), metadata=metadata)
            >>> ec.metadata  # doctest: +NORMALIZE_WHITESPACE
            {'figure description': {'fields': [{'name': 't', 'unit': 's'},
            {'name': 'E', 'unit': 'V', 'reference': 'RHE'},
            {'name': 'j', 'unit': 'uA / cm2'}, {'name': 'x', 'unit': 'm'}]}}

        """
        return self._metadata or self.loader.metadata



r"""
The echemdb-converter suite.

EXAMPLES::

    >>> from echemdbconverters.test.cli import invoke
    >>> invoke(cli, "--help")  # doctest: +NORMALIZE_WHITESPACE
    Usage: cli [OPTIONS] COMMAND [ARGS]...

      The echemdb-converter suite

    Options:
      --help  Show this message and exit.
    Commands:
      csv  Convert a file containing CSV data into an echemdb unitpackage.
      ec   Convert an electrochemistry file into an echemdb datapackage.

"""
# ********************************************************************
#  This file is part of echemdb-converters.
#
#        Copyright (C) 2022 Albert Engstfeld
#        Copyright (C) 2022 Johannes Hermann
#        Copyright (C) 2022 Julian Rüth
#
#  echemdb-converters is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  echemdb-converters is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with echemdb-converters. If not, see <https://www.gnu.org/licenses/>.
# ********************************************************************
import logging
import os

import click

logger = logging.getLogger("echemdb-converters")


@click.group(help=__doc__.split("EXAMPLES")[0])
def cli():
    r"""
    Entry point of the command line interface.

    This redirects to the individual commands listed below.
    """


def _outfile(template, suffix=None, outdir=None):
    r"""
    Return a file name for writing.

    The file is named like `template` but with the suffix changed to `suffix`
    if specified. The file is created in `outdir`, if specified, otherwise in
    the directory of `template`.

    EXAMPLES::

        >>> from echemdbconverters.test.cli import invoke, TemporaryData
        >>> with TemporaryData("../**/default.cvs") as directory:
        ...     outname = _outfile(os.path.join(directory, "default.csv"), suffix=".csv")
        ...     with open(outname, mode="wb") as csv:
        ...         _ = csv.write(b"...")
        ...     os.path.exists(os.path.join(directory, "default.csv"))
        True

    ::

        >>> with TemporaryData("../**/default.csv") as directory:
        ...     outname = _outfile(os.path.join(directory, "default.csv"), suffix=".csv", outdir=os.path.join(directory, "subdirectory"))
        ...     with open(outname, mode="wb") as csv:
        ...         _ = csv.write(b"...")
        ...     os.path.exists(os.path.join(directory, "subdirectory", "default.csv"))
        True

    """
    if suffix is not None:
        template = f"{os.path.splitext(template)[0]}{suffix}"

    if outdir is not None:
        template = os.path.join(outdir, os.path.basename(template))

    os.makedirs(os.path.dirname(template) or ".", exist_ok=True)

    return template


def _create_package(csvname, metadata, fields, outdir):
    r"""
    Return a data package built from a :param:`metadata` dict and tabular data
    in :param:`csvname`.

    This is a helper method for :meth:`convert`.
    """
    from frictionless import Package, Resource, Schema

    package = Package(
        resources=[
            Resource(
                path=os.path.basename(csvname),
                basepath=outdir or os.path.dirname(csvname),
            )
        ],
    )
    package.infer()

    resource = package.resources[0]

    resource.custom.setdefault("metadata", {})
    resource.custom["metadata"].setdefault("echemdb", metadata)

    resource.schema = Schema(fields=fields)

    return package


def _write_metadata(out, metadata):
    r"""
    Write `metadata` to the `out` stream in JSON format.

    This is a helper method for :meth:`digitize_cv`.
    """

    def defaultconverter(item):
        r"""
        Return `item` that Python's json package does not know how to serialize
        in a format that Python's json package does know how to serialize.
        """
        from datetime import date, datetime

        # The YAML standard knows about dates and times, so we might see these
        # in the metadata. However, standard JSON does not know about these so
        # we need to serialize them as strings explicitly.
        if isinstance(item, (datetime, date)):
            return str(item)

        raise TypeError(f"Cannot serialize ${item} of type ${type(item)} to JSON.")

    import json

    json.dump(metadata, out, default=defaultconverter, ensure_ascii=False, indent=4)
    # json.dump does not save files with a newline, which compromises the tests
    # where the output files are compared to an expected json.
    out.write("\n")


@click.command(name="csv")
@click.argument("csv", type=click.Path(exists=True))
@click.option("--device", type=str, default=None, help="selects a specific CSVloader")
@click.option(
    "--outdir",
    type=click.Path(file_okay=False),
    default=None,
    help="write output files to this directory",
)
@click.option(
    "--metadata", type=click.File("rb"), default=None, help="yaml file with metadata"
)
def convert(csv, device, outdir, metadata):
    """
    Convert a file containing CSV data into an echemdb unitpackage.
    \f

    EXAMPLES::

        >>> from echemdbconverters.test.cli import invoke, TemporaryData
        >>> with TemporaryData("../**/default.csv") as directory:
        ...     invoke(cli, "csv", os.path.join(directory, "default.csv"))

    TESTS:

    The command can be invoked on files in the current directory::

        >>> from echemdbconverters.test.cli import invoke, TemporaryData
        >>> cwd = os.getcwd()
        >>> with TemporaryData("../**/default.csv") as directory:
        ...     os.chdir(directory)
        ...     try:
        ...         invoke(cli, "csv", "default.csv")
        ...     finally:
        ...         os.chdir(cwd)

    """
    import yaml

    from echemdbconverters.csvloader import CSVloader

    if device:
        with open(csv, "r") as file:  # pylint: disable=unspecified-encoding
            loader = CSVloader.create(device)(file)
    else:
        with open(csv, "r") as file:  # pylint: disable=unspecified-encoding
            loader = CSVloader(file)

    if metadata:
        metadata = yaml.load(metadata, Loader=yaml.SafeLoader)

    fields = loader.derive_fields()

    _create_outfiles(csv, loader, fields, metadata, outdir)


def _create_outfiles(csv, loader, fields, metadata, outdir):
    # write new csv
    csvname = _outfile(csv, suffix=".csv", outdir=outdir)
    loader.df.to_csv(csvname, index=False)

    # write package
    package = _create_package(csvname, metadata, fields, outdir)

    with open(
        _outfile(csv, suffix=".json", outdir=outdir),
        mode="w",
        encoding="utf-8",
    ) as json:
        _write_metadata(json, package.to_dict())


@click.command(name="ec")
@click.argument("csv", type=click.Path(exists=True))
@click.option("--device", type=str, default=None, help="selects a specific CSVloader")
@click.option(
    "--outdir",
    type=click.Path(file_okay=False),
    default=None,
    help="write output files to this directory",
)
@click.option(
    "--metadata", type=click.File("rb"), default=None, help="yaml file with metadata"
)
def electrochemistry(csv, device, outdir, metadata):
    r"""
    Convert an electrochemistry file into an echemdb datapackage.

    EXAMPLES::

        >>> from echemdbconverters.test.cli import invoke, TemporaryData
        >>> with TemporaryData("unit.csv") as directory:
        ...     with TemporaryData("unit.csv.metadata") as directory2:
        ...         invoke(cli, "ec", os.path.join(directory, "unit.csv"), '--outdir', os.path.join(directory, "generated"), '--metadata', os.path.join(directory2, "unit.csv.metadata"))

    TESTS:

    The command can be invoked on files in the current directory::

        >>> from echemdbconverters.test.cli import invoke, TemporaryData
        >>> cwd = os.getcwd()
        >>> with TemporaryData("unit.csv") as directory:
        ...     with TemporaryData("unit.csv.metadata") as directory2:
        ...         os.chdir(directory)
        ...         try:
        ...             invoke(cli, "ec", os.path.join(directory, "unit.csv"), '--outdir', os.path.join(directory, "generated"), '--metadata', os.path.join(directory2, "unit.csv.metadata"))
        ...         finally:
        ...             os.chdir(cwd)

    """
    import yaml

    from echemdbconverters.csvloader import CSVloader
    from echemdbconverters.ec_unit_package_adapter import ECUnitPackageAdapter

    fields = None

    if metadata:
        metadata = yaml.load(metadata, Loader=yaml.SafeLoader)
        try:
            fields = metadata["figure description"]["schema"]["fields"]
        except (KeyError, AttributeError):
            logger.warning("No units to the fields provided in the metadata")

    if device:
        with open(csv, "r") as file:  # pylint: disable=unspecified-encoding
            loader = ECUnitPackageAdapter.create(device=device)(
                CSVloader.create(device)(file), fields=fields
            )
    else:
        with open(csv, "r") as file:  # pylint: disable=unspecified-encoding
            loader = ECUnitPackageAdapter(CSVloader(file), fields=fields)

    fields = loader.fields()

    _create_outfiles(csv, loader, fields, metadata, outdir)


cli.add_command(convert)
cli.add_command(electrochemistry)

# Register command docstrings for doctesting.
# Since commands are not functions anymore due to their decorator, their
# docstrings would otherwise be ignored.
__test__ = {
    name: command.__doc__ for (name, command) in cli.commands.items() if command.__doc__
}



r"""
Helpers for click CLI testing.

Click's own CliRunner is quite cumbersome to work with in some simple test
scenarios so we wrap it in more convenient ways here.

The code has been inherited from svgdigitizer and adapted for echemdb-converters.

"""
# *********************************************************************
#  This file is part of svgdigitizer.
#
#        Copyright (C) 2021 Julian Rüth
#
#  svgdigitizer is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  svgdigitizer is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with svgdigitizer. If not, see <https://www.gnu.org/licenses/>.
# *********************************************************************


def invoke(command, *args):
    r"""
    Invoke the click ``command`` with the given list of string arguments.

    >>> import click
    >>> @click.command()
    ... def hello(): print("Hello World")
    >>> invoke(hello)
    Hello World

    >>> @click.command()
    ... def fails(): raise Exception("expected error")
    >>> invoke(fails)
    Traceback (most recent call last):
    ...
    Exception: expected error

    """
    from click.testing import CliRunner

    invocation = CliRunner().invoke(command, args, catch_exceptions=False)
    output = invocation.output.strip()
    if output:
        print(output)


class TemporaryData:
    r"""
    Provides a temporary directory with test files.

    EXAMPLES::

        >>> import os
        >>> with TemporaryData("unit.*") as directory:
        ...     'unit.csv' in os.listdir(directory)
        True

    """

    def __init__(self, *patterns):
        self._patterns = patterns
        self._tmpdir = None

    def __enter__(self):
        import tempfile

        self._tmpdir = tempfile.TemporaryDirectory()

        try:
            import glob
            import os
            import os.path
            import shutil

            import echemdbconverters

            cwd = os.getcwd()
            os.chdir(
                os.path.join(
                    os.path.dirname(echemdbconverters.__file__), "..", "test", "csv"
                )
            )
            try:
                for pattern in self._patterns:
                    for filename in glob.glob(pattern):
                        shutil.copy(filename, self._tmpdir.name)

                return self._tmpdir.name
            finally:
                os.chdir(cwd)
        except Exception:
            self._tmpdir.cleanup()
            raise

    def __exit__(self, *args):
        self._tmpdir.cleanup()



r"""
Loader for CSV files (https://datatracker.ietf.org/doc/html/rfc4180)
which consist of a single header line containing the column (field)
names and rows with comma separated values.

In pandas the names of the columns are referred to as `column_names`,
whereas in a frictionless datapackage the column names are called `fields`.
The datapackage contains information about, i.e.,
the type of data, a title and a set of descriptors.

The CSV object has the following properties:

TODO:: Add examples for the following functions
    * a DataFrame
    * the column names
    * the header contents
    * the number of header lines

Loaders for non standard CSV files can be called:

TODO:: Add example

"""
# ********************************************************************
#  This file is part of echemdb-converters.
#
#        Copyright (C) 2022 Albert Engstfeld
#        Copyright (C) 2022 Johannes Hermann
#        Copyright (C) 2022 Julian Rüth
#
#  echemdb-converters is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  echemdb-converters is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with echemdb-converters. If not, see <https://www.gnu.org/licenses/>.
# ********************************************************************


import logging

logger = logging.getLogger("loader")


class CSVloader:
    r"""
    Loads a CSV, where the first line must contain the column (field) names
    and the following lines comma separated values.

    EXAMPLES::

        >>> from io import StringIO
        >>> file = StringIO(r'''a,b
        ... 0,0
        ... 1,1''')
        >>> csv = CSVloader(file)
        >>> csv.df
           a  b
        0  0  0
        1  1  1

    A list of column names::

        >>> csv.column_names
        ['a', 'b']

    More specific converters can be selected:

    TODO:: Add example with csv.get_device('device')(file)

    """

    def __init__(self, file):
        self._file = file.read()

    @property
    def file(self):
        r"""
        A file like object of the loaded CSV.

        EXAMPLES::
            >>> from io import StringIO
            >>> file = StringIO(r'''a,b
            ... 0,0
            ... 1,1''')
            >>> csv = CSVloader(file)
            >>> type(csv.file)
            <class '_io.StringIO'>

        """
        from io import StringIO

        return StringIO(self._file)

    def augment(self, metadata=None):
        r"""
        Metadata constructed from input metadata and the CSV header.
        A simple CSV does not have any metadata in the header.

        EXAMPLES:

        Without metadata::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j
            ... 0,0,0
            ... 1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> csv = CSVloader(file)
            >>> csv.augment()
            {}

        Without metadata provided to the loader::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j
            ... 0,0,0
            ... 1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> csv = CSVloader(file)
            >>> csv.augment(metadata={'foo':'bar'})
            {'foo': 'bar'}

        """
        return metadata or {}

    @staticmethod
    def create(device=None):
        r"""
        Calls a specific `loader` based on a given device.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO('''EC-Lab ASCII FILE
            ... Nb header lines : 6
            ...
            ... Device metadata : some metadata
            ...
            ... mode\ttime/s\tEwe/V\t<I>/mA\tcontrol/V
            ... 2\t0\t0.1\t0\t0
            ... 2\t1\t1.4\t5\t1
            ... ''')
            >>> csv = CSVloader.create('eclab')(file)
            >>> csv.df
            mode  time/s Ewe/V  <I>/mA  control/V
            0     2       0   0.1       0          0
            1     2       1   1.4       5          1

        """
        if device == "eclab":
            from echemdbconverters.eclabloader import ECLabLoader

            return ECLabLoader

        raise KeyError(f"Device wth name '{device}' is unknown to the loader'.")

    @property
    def df(self):
        r"""
        A pandas dataframe of the data in the CSV.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''a,b
            ... 0,0
            ... 1,1''')
            >>> csv = CSVloader(file)
            >>> csv.df
               a  b
            0  0  0
            1  1  1

        """
        import pandas as pd

        return pd.read_csv(self.file, header=self.header_lines)

    @property
    def header(self):
        r"""
        The header of the CSV (column names excluded).

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''a,b
            ... 0,0
            ... 1,1''')
            >>> csv = CSVloader(file)
            >>> csv.header
            []

        """
        lines = self.file.readlines()

        return [lines[_] for _ in range(self.header_lines)]

    @property
    def data(self):
        r"""
        A file like object with the data of the CSV without header lines.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''a,b
            ... 0,0
            ... 1,1''')
            >>> csv = CSVloader(file)
            >>> type(csv.data)
            <class '_io.StringIO'>

            >>> from io import StringIO
            >>> file = StringIO(r'''a,b
            ... 0,0
            ... 1,1''')
            >>> csv = CSVloader(file)
            >>> csv.data.readlines()
            ['0,0\n', '1,1']

        """
        from io import StringIO

        return StringIO(
            "".join(line for line in self.file.readlines()[self.header_lines + 1 :])
        )

    @property
    def column_names(self):
        r"""
        List of column (field) names describing the tabulated data.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''a,b
            ... 0,0
            ... 1,1''')
            >>> csv = CSVloader(file)
            >>> csv.column_names
            ['a', 'b']

        """
        return self.df.columns.to_list()

    @property
    def header_lines(self):
        r"""
        The number of header lines in a CSV excluding the line with the column names.

        EXAMPLES:

        Files for the base loader do not have a header::

            >>> from io import StringIO
            >>> file = StringIO(r'''a,b
            ... 0,0
            ... 1,1''')
            >>> csv = CSVloader(file)
            >>> csv.header_lines
            0

        Implementation in a specific device loader::

            >>> file = StringIO('''EC-Lab ASCII FILE
            ... Nb header lines : 6
            ...
            ... Device metadata : some metadata
            ...
            ... mode\ttime/s\tEwe/V\t<I>/mA\tcontrol/V
            ... 2\t0\t0,1\t0\t0
            ... 2\t1\t1,4\t5\t1
            ... ''')
            >>> csv = CSVloader.create('eclab')(file)
            >>> csv.header_lines
            5

        """
        return 0

    def derive_fields(self, fields=None):
        r"""
        Fields describing the column names.

        EXAMPLES:

        If not specified, the fields are created from the `column_names`.::

            >>> from io import StringIO
            >>> file = StringIO(r'''t,E,j
            ... 0,0,0
            ... 1,1,1''')
            >>> from echemdbconverters.csvloader import CSVloader
            >>> csv = CSVloader(file)
            >>> csv.derive_fields()
            [{'name': 't', 'type': 'integer'}, {'name': 'E', 'type': 'integer'}, {'name': 'j', 'type': 'integer'}]

        The fields can be provided as an argument to the loader.::

            >>> file = StringIO(r'''t,E,j
            ... 0,0,0
            ... 1,1,1''')
            >>> metadata = {'figure description': {'schema': {'fields': [{'name':'t', 'unit':'s'},{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'}]}}}
            >>> csv = CSVloader(file=file)
            >>> csv.derive_fields(fields=metadata['figure description']['schema']['fields'])
            [{'name': 't', 'type': 'integer', 'unit': 's'}, {'name': 'E', 'type': 'integer', 'unit': 'V', 'reference': 'RHE'}, {'name': 'j', 'type': 'integer', 'unit': 'uA / cm2'}]

        When a field is missing (here `t`) it will be generated and all obsolete fields are removed.::

            >>> file = StringIO(r'''t,E,j
            ... 0,0,0
            ... 1,1,1''')
            >>> metadata = {'figure description': {'schema': {'fields': [{'name':'E', 'unit':'V', 'reference':'RHE'},{'name':'j', 'unit':'uA / cm2'},{'name': 'x'},{'foo':'bar'}]}}}
            >>> csv = CSVloader(file=file)
            >>> csv.derive_fields(fields=metadata['figure description']['schema']['fields'])
            [{'name': 't', 'type': 'integer'}, {'name': 'E', 'type': 'integer', 'unit': 'V', 'reference': 'RHE'}, {'name': 'j', 'type': 'integer', 'unit': 'uA / cm2'}]

        """
        from frictionless import Resource, Schema

        # infer fields from pandas dataframe
        df_resource = Resource(self.df)
        df_resource.infer()

        if fields:
            # validate that fields have a name
            _fields = []

            def validate_field(field):
                try:
                    field["name"]
                except KeyError:
                    return False
                return True

            for field in fields:
                if validate_field(field):
                    _fields.append(field)

            schema = Schema.from_descriptor({"fields": _fields}, allow_invalid=True)

            # Update the df_resource schema with additional information from the metadata
            for name in schema.field_names:
                if name in self.column_names:
                    field = schema.get_field(name).to_dict()
                    for key in field:
                        if not key == "name":
                            df_resource.schema.update_field(name, {key: field[key]})

        return df_resource.schema.fields

    @property
    def delimiter(self):
        r"""
        The delimiter in the CSV, which is extracted from
        the first two lines of the CSV data.

        EXAMPLES::

            >>> from io import StringIO
            >>> file = StringIO(r'''a,b
            ... 0,0
            ... 1,1''')
            >>> csv = CSVloader(file)
            >>> csv.delimiter
            ','

        """
        import clevercsv

        # Only two lines are used to detect the delimiter,
        # since clevercsv is slow with files containing many columns.
        return clevercsv.detect.Detector().detect(self.data.read()[:2]).delimiter

    @property
    def decimal(self):
        r"""
        The decimal separator in the floats in the CSV data.

        EXAMPLES:

        Not implemented in the base loader::

            >>> from io import StringIO
            >>> file = StringIO(r'''a,b
            ... 0,0
            ... 1,1''')
            >>> csv = CSVloader(file)
            >>> csv.decimal
            Traceback (most recent call last):
            ...
            NotImplementedError

        Implementation in a specific device loader::

            >>> file = StringIO('''EC-Lab ASCII FILE
            ... Nb header lines : 6
            ...
            ... Device metadata : some metadata
            ...
            ... mode\ttime/s\tEwe/V\t<I>/mA\tcontrol/V
            ... 2\t0\t0,1\t0\t0
            ... 2\t1\t1,4\t5\t1
            ... ''')
            >>> csv = CSVloader.create('eclab')(file)
            >>> csv.decimal
            ','

        """
        raise NotImplementedError



# *********************************************************************
#  This file is part of echemdb-converters.
#
#        Copyright (C) 2023 Albert Engstfeld
#
#  echemdb-converters is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  echemdb-converters is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with echemdb-converters. If not, see <https://www.gnu.org/licenses/>.
# *********************************************************************
from distutils.core import setup

setup(
    name="echemdbconverters",
    version="0.1.0",
    packages=["echemdbconverters", "echemdbconverters.test"],
    license="GPL 3.0+",
    description="echemdb-converters is a Python library and command line tool to load csv data and convert those into frictionless datapackages.",
    long_description=open("README.md", encoding="UTF-8").read(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=[
        "clevercsv",
        "click>=8,<9",
        "frictionless>=5,<6",
        "pandas",
    ],
    entry_points={
        "console_scripts": ["echemdbconverters=echemdbconverters.entrypoint:cli"],
    },
    python_requires=">=3.9",
)
